'''Train scNym models and identify cell type markers'''
import numpy as np
import pandas as pd
from scipy import sparse
import os
import os.path as osp
import scanpy as sc

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import DataLoader
from sklearn.model_selection import StratifiedKFold
from typing import Union
import copy
import itertools
from functools import partial

from .model import CellTypeCLF, CellTypeCLFConditional

from .dataprep import SingleCellDS, SampleMixUp, balance_classes
from .dataprep import AUGMENTATION_SCHEMES
from .trainer import Trainer, SemiSupervisedTrainer
from .trainer import cross_entropy, get_class_weight
from .trainer import InterpolationConsistencyLoss, ICLWeight, MixMatchLoss, DANLoss

from .predict import Predicter
from . import utils

# define optimizer map for cli selection
OPTIMIZERS = {
    'adadelta': torch.optim.Adadelta,
    'adam': torch.optim.Adam,
    'adamw': torch.optim.AdamW,
    'sgd': torch.optim.SGD,
}

#########################################################
# Train scNym classification models
#########################################################


def repeater(data_loader):
    """Use `itertools.repeat` to infinitely loop through
    a dataloader.
    
    Parameters
    ----------
    data_loader : torch.utils.data.DataLoader
        data loader class.
    
    Yields
    ------
    data : Iterable
        batches from `data_loader`.
        
    Credit
    ------
    https://bit.ly/2z0LGm8
    """
    for loader in itertools.repeat(data_loader):
        for data in loader:
            yield data


def fit_model(
    X: Union[np.ndarray, sparse.csr.csr_matrix],
    y: np.ndarray,
    traintest_idx: Union[np.ndarray, tuple],
    val_idx: np.ndarray,
    batch_size: int,
    n_epochs: int,
    lr: float,
    optimizer_name: str,
    weight_decay: float,
    ModelClass: nn.Module,
    out_path: str,
    n_genes: int = None,
    mixup_alpha: float = None,
    unlabeled_counts: np.ndarray = None,
    unsup_max_weight: float = 2.,
    unsup_mean_teacher: bool = False,
    ssl_method: str='mixmatch',
    ssl_kwargs: dict={},
    weighted_classes: bool = False,
    balanced_classes: bool = False,
    input_domain: np.ndarray=None,
    unlabeled_domain: np.ndarray=None,
    pretrained: str=None,
    patience: int=None,
    save_freq: int=None,
    tensorboard: bool=True,
    **kwargs,
) -> (float, float):
    '''Fit an scNym model given a set of observations and labels.
    
    Parameters
    ----------
    X : np.ndarray
        [Cells, Genes] of log1p transformed normalized values.
        log1p and normalization performed using scanpy defaults.
    y : np.ndarray
        [Cells,] integer class labels.
    traintest_idx : np.ndarray
        [Int,] indices to use for training and early stopping.
        a single array will be randomly partitioned, OR a tuple
        of `(train_idx, test_idx)` can be passed.
    val_idx : np.ndarray
        [Int,] indices to hold-out for final model evaluation.
    n_epochs : int
        number of epochs for training.
    lr : float
        learning rate.
    optimizer_name : str
        optimizer to use. {"adadelta", "adam"}.
    weight_decay : float
        weight decay to apply to model weights.
    ModelClass : nn.Module
        a model class for construction classification models.
    batch_size : int
        batch size for training.        
    fold_indices : list
        elements are 2-tuple, with training indices and held-out.
    out_path : str
        top level path for saving fold outputs.
    n_genes : int
        number of genes in the input. Not necessarily `X.shape[1]` if
        the input matrix has been concatenated with other features.
    mixup_alpha : float
        alpha parameter for an optional MixUp augmentation during training.
    unlabeled_counts : np.ndarray
        [Cells', Genes] of log1p transformed normalized values for 
        unlabeled observations.
    unsup_max_weight : float
        maximum weight for the unsupervised loss term.
    unsup_mean_teacher : bool
        use a mean teacher for pseudolabel generation.
    ssl_method : str
        semi-supervised learning method to use.
    ssl_kwargs : dict
        arguments passed to the semi-supervised learning loss.
    balanced_classes : bool
        perform class balancing by undersampling majority classes.
    weighted_classes : bool
        weight loss for each class based on relative abundance of classes
        in the training data.
    input_domain : np.ndarray
        [Cells,] integer domain labels for training data.
    unlabeled_domain : np.ndarray
        [Cells',] integer domain labels for unlabeled data.
    pretrained : str
        path to a pretrained model for initialization.
        default: `None`.
    patience : int
        number of epochs to wait before early stopping.
        `None` deactivates early stopping.
    save_freq : int
        frequency in epochs for saving model checkpoints.
        if `None`, saves >=5 checkpoints per model.
    tensorboard : bool
        save logs to tensorboard.
        
    Returns
    -------
    test_acc : float
        classification accuracy on the test set.
    test_loss : float
        supervised loss on the test set.
    '''
    # count the number of cell types available
    n_cell_types = len(np.unique(y))
    if n_genes is None:
        n_genes = X.shape[1]
        
    if type(traintest_idx) != tuple:
        # Set aside 10% of the traintest data for model selection in `test_idx`
        train_idx = np.random.choice(
            traintest_idx,
             size=int(
                 np.floor(0.9 * len(traintest_idx))),
             replace=False,
        ).astype('int')
        test_idx = np.setdiff1d(traintest_idx, train_idx).astype('int')
    elif type(traintest_idx) == tuple and len(traintest_idx) == 2:
        # use the user provided train/test split
        train_idx = traintest_idx[0]
        test_idx  = traintest_idx[1]
    else:
        # the user supplied an invalid argument
        msg = '`traintest_idx` of type {type(traintest_idx)}\n'
        msg += 'and length {len(traintest_idx)} is invalid.'
        raise ValueError(msg)
        
    # save indices to CSVs for later retrieval
    np.savetxt(osp.join(out_path, 'train_idx.csv'), train_idx)
    np.savetxt(osp.join(out_path, 'test_idx.csv'), test_idx)
    np.savetxt(osp.join(out_path, 'val_idx.csv'), val_idx)

    # balance or weight classes if applicable
    if balanced_classes and weighted_classes:
        msg = 'balancing AND weighting classes is not useful.'
        msg += '\nPick one mode of accounting for class imbalances.'
        raise ValueError(msg)
    elif balanced_classes and not weighted_classes:
        print('Setting up a stratified sampler...')
        # we sample classes with weighted likelihood, rather than
        # a uniform likelihood of sampling
        # we use the inverse of the class count as a weight
        # this is normalized in `WeightedRandomSample`
        classes, counts = np.unique(y[train_idx], return_counts=True)
        sample_weights = 1./counts
        
        # `WeightedRandomSampler` is kind of funny and takes a weight
        # **per example** in the training set, rather than per class.
        # here we assign the appropriate class weight to each sample
        # in the training set.
        weight_per_example = sample_weights[y[train_idx]]
        
        # we instantiate the sampler with the relevant weight for
        # each observation and set the number of total samples to the
        # number of samples in our training set
        # `WeightedRandomSampler` will sample indices from a multinomial
        # with probabilities computed from the normalized vector
        # of `weights_per_example`.
        sampler = torch.utils.data.sampler.WeightedRandomSampler(
            weight_per_example, 
            len(y[train_idx]),
        )
        class_weight = None
    elif weighted_classes and not balanced_classes:
        # compute class weights
        # class weights amplify the loss of some classes and reduce
        # the loss of others, inversely proportional to the class
        # frequency
        print('Weighting classes for training...')
        class_weight = get_class_weight(y[train_idx])
        print(class_weight)
        print()
        sampler = None
    else:
        print('Not weighting classes and not balancing classes.')
        class_weight = None
        sampler = None

    # Generate training and model selection Datasets and Dataloaders
    X_train = X[train_idx, :]
    y_train = y[train_idx]

    X_test = X[test_idx, :]
    y_test = y[test_idx]
    
    # count the number of domains
    if (input_domain is None) and (unlabeled_domain is None):
        n_domains = 2
    elif (input_domain is not None) and (unlabeled_domain is not None):
        n_domains = int(np.max([
            input_domain.max(), unlabeled_domain.max(),
        ])) + 1
    else:
        msg = 'domains supplied for only one set of data'
        raise ValueError(msg)
    print(f'Found {n_domains} unique domains.')
    
    if input_domain is not None:
        d_train = input_domain[train_idx]
        d_test  = input_domain[test_idx]
    else:
        d_train = None
        d_test  = None

    train_ds = SingleCellDS(
        X=X_train,
        y=y_train,
        num_classes=len(np.unique(y)),
        domain = d_train,
        num_domains = n_domains,
    )
    test_ds = SingleCellDS(
        X_test,
        y_test,
        num_classes=len(np.unique(y)),
        domain = d_test,
        num_domains = n_domains,
    )

    train_dl = DataLoader(
        train_ds,
        batch_size=batch_size,
        shuffle=True if sampler is None else False,
        sampler=sampler,
        drop_last=True,
    )
    test_dl = DataLoader(
        test_ds,
        batch_size=batch_size,
        shuffle=True,
    )

    dataloaders = {
        'train': train_dl,
        'val': test_dl,
    }

    # Define batch transformers
    batch_transformers = {}
    if mixup_alpha is not None and ssl_method != 'mixmatch':
        print('Using MixUp as a batch transformer.')
        batch_transformers['train'] = SampleMixUp(alpha=mixup_alpha)

    # Build a cell type classification model and transfer to CUDA
    model = ModelClass(
        n_genes=n_genes,
        n_cell_types=n_cell_types,
        **kwargs,
    )
    
    if pretrained is not None:
        # initialize with supplied weights
        model.load_state_dict(
            torch.load(
                pretrained,
                map_location='cpu',
            )
        )
    
    if torch.cuda.is_available():
        model = model.cuda()

    # Set up loss criterion and the model optimizer
    # here we use our own cross_entropy loss to handle
    # discrete probability distributions rather than
    # categorical predictions
    if class_weight is None:
        criterion = cross_entropy
    else:
        criterion = partial(
            cross_entropy,
            class_weight=torch.from_numpy(class_weight).float(),
        )
        
    opt_callable = OPTIMIZERS[optimizer_name.lower()]
    
    if opt_callable != torch.optim.SGD:
        optimizer = opt_callable(
            model.parameters(),
            weight_decay=weight_decay,
            lr=lr,
        )
        scheduler = None
    else:
        # use SGD as the optimizer with momentum
        # and a learning rate scheduler
        optimizer = opt_callable(
            model.parameters(),
            weight_decay=weight_decay,
            lr=lr,
            momentum=0.9,
        )
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
            optimizer=optimizer,
            T_max=n_epochs,
            eta_min=lr/10000,
        )
        

    # Build the relevant trainer object for either supervised
    # or semi-supervised learning with interpolation consistency
    trainer_kwargs = {
        'model': model,
        'criterion': criterion,
        'optimizer': optimizer,
        'scheduler': scheduler,
        'dataloaders': dataloaders,
        'out_path': out_path,
        'batch_transformers': batch_transformers,
        'n_epochs': n_epochs,
        'min_epochs': n_epochs//20,
        'save_freq': max(n_epochs//5, 1) if save_freq is None else save_freq,
        'reg_criterion': None,
        'exp_name': osp.basename(out_path),
        'verbose': False,
        'tb_writer' : osp.join(out_path, 'tblog') if tensorboard else None,
        'patience': patience,
    }

    if unlabeled_counts is None:
        # perform fully supervised training
        T = Trainer(**trainer_kwargs)
    else:
        # perform semi-supervised training
        unsup_dataset = SingleCellDS(
            X=unlabeled_counts,
            y=np.zeros(unlabeled_counts.shape[0]),
            num_classes=len(np.unique(y)),
            domain=unlabeled_domain,
            num_domains=n_domains,
        )
        
        # Build a semi-supervised data loader that infinitely samples
        # unsupervised data for interpolation consistency.
        # This allows us to loop through the labeled data iterator
        # without running out of unlabeled batches.        
        unsup_dataloader = DataLoader(
            unsup_dataset,
            batch_size=batch_size,
            shuffle=True,
            drop_last=True,
        )
        unsup_dataloader = repeater(unsup_dataloader)

        # Set up the unsupervised loss
        if ssl_method.lower() == 'ict':
            print('Using ICT for semi-supervised learning')                
            USL = InterpolationConsistencyLoss(
                alpha=mixup_alpha if mixup_alpha is not None else 0.3,
                unsup_criterion=nn.MSELoss(),
                sup_criterion=criterion,
                decay_coef=ssl_kwargs.get('decay_coef', 0.997),
                mean_teacher=unsup_mean_teacher,
            )

        elif ssl_method.lower() == 'mixmatch':
            print('Using MixMatch for semi-supervised learning')
            # we want the raw MSE per sample here, rather than the average
            # so we set `reduction='none'`.
            # this allows us to scale the weight of individual examples
            # based on pseudolabel confidence.
            unsup_criterion_name = ssl_kwargs.get('unsup_criterion', 'mse')
            if unsup_criterion_name.lower() == 'mse':
                unsup_criterion = nn.MSELoss(reduction='none')
            elif unsup_criterion_name.lower() in ('crossentropy', 'ce'):
                unsup_criterion = partial(
                    cross_entropy,
                    reduction='none',
                )
            USL = MixMatchLoss(
                alpha=mixup_alpha if mixup_alpha is not None else 0.3,
                unsup_criterion=unsup_criterion,
                sup_criterion=criterion,
                decay_coef=ssl_kwargs.get('decay_coef', 0.997),
                mean_teacher=unsup_mean_teacher,
                augment=AUGMENTATION_SCHEMES[ssl_kwargs.get('augment', 'log1p_drop')],
                n_augmentations=ssl_kwargs.get('n_augmentations', 1),
                T=ssl_kwargs.get('T', 0.5),
                augment_pseudolabels=ssl_kwargs.get('augment_pseudolabels', True),
                pseudolabel_min_confidence=ssl_kwargs.get('pseudolabel_min_confidence', 0.0),
            )
        else:
            msg = f'{ssl_method} is not a valid semi-supervised learning method.\n'
            msg += 'must be one of {"ict", "mixmatch"}'
            raise ValueError(msg)

        # set up the weight schedule
        # we define a number of epochs for ramping, a number to wait
        # ("burn_in_epochs") before we start the ramp up, and a maximum
        # coefficient value
        weight_schedule = ICLWeight(
            ramp_epochs=ssl_kwargs.get('ramp_epochs', max(n_epochs//4, 1)),
            max_unsup_weight=unsup_max_weight,
            burn_in_epochs = ssl_kwargs.get('burn_in_epochs', 20),
            sigmoid = ssl_kwargs.get('sigmoid', False),
        )
        # don't let early stopping save checkpoints from before the SSL 
        # ramp up has started
        trainer_kwargs['min_epochs'] = max(
            trainer_kwargs['min_epochs'],
            weight_schedule.burn_in_epochs + weight_schedule.ramp_epochs // 5,
        )
        
        # if min_epochs are manually specified, use that number instead
        if ssl_kwargs.get('min_epochs', None) is not None:
            trainer_kwargs['min_epochs'] = ssl_kwargs['min_epochs']
        
        # let the model save weights even if the ramp is 
        # longer than the total epochs we'll train for
        trainer_kwargs['min_epochs'] = min(
            trainer_kwargs['min_epochs'],
            trainer_kwargs['n_epochs']-1,
        )
        
        dan_criterion = ssl_kwargs.get('dan_criterion', None)
        if dan_criterion is not None:
            # initialize the DAN Loss
            
            dan_criterion = DANLoss(
                model=model,
                dan_criterion=cross_entropy,
                use_conf_pseudolabels=ssl_kwargs.get('dan_use_conf_pseudolabels', False),
                scale_loss_pseudoconf=ssl_kwargs.get('dan_scale_loss_pseudoconf', False),
                n_domains = n_domains,
            )

            # setup the DANN learning rate schedule
            dan_weight = ICLWeight(
                ramp_epochs=ssl_kwargs.get('dan_ramp_epochs', max(n_epochs//4, 1)),
                max_unsup_weight=ssl_kwargs.get('dan_max_weight', 1.),
                burn_in_epochs = ssl_kwargs.get('dan_burn_in_epochs', 0),
                sigmoid = ssl_kwargs.get('sigmoid', True),
            )
            # add DANN parameters to the optimizer
            optimizer.add_param_group({
                'params': dan_criterion.dann.domain_clf.parameters(),
                'name': 'domain_classifier',
            })
        else:
            dan_weight = None

        # initialize the trainer
        T = SemiSupervisedTrainer(
            unsup_dataloader=unsup_dataloader,
            unsup_criterion=USL,
            unsup_weight=weight_schedule,
            dan_criterion=dan_criterion,
            dan_weight=dan_weight,
            **trainer_kwargs,
        )

    print('Training...')
    T.train()
    print('Training complete.')
    print()

    # Perform model evaluation using the best set of weights on the
    # totally unseen, held out data.
    print('Evaluating model.')
    model = ModelClass(
        n_genes=n_genes,
        n_cell_types=n_cell_types,
        **kwargs,
    )
    model.load_state_dict(
        torch.load(
            osp.join(out_path, '00_best_model_weights.pkl'),
        )
    )
    model.eval()

    if torch.cuda.is_available():
        model = model.cuda()

    # Build a DataLoader for validation
    X_val = X[val_idx, :]
    y_val = y[val_idx]
    val_ds = SingleCellDS(
        X_val, 
        y_val,
        num_classes=len(np.unique(y)),            
    )
    val_dl = DataLoader(
        val_ds, 
        batch_size=batch_size, 
        shuffle=False, 
    )

    # Without recording any gradients to speed things up,
    # predict classes for all held out data and evaluate metrics.
    with torch.no_grad():
        loss = 0.
        running_corrects = 0.
        running_total = 0.
        all_predictions = []
        all_labels = []
        for data in val_dl:
            input_ = data['input']

            label_ = data['output']  # one-hot

            if torch.cuda.is_available():
                input_ = input_.cuda()
                label_ = label_.cuda()

            # make an integer version of labels for convenience
            int_label_ = torch.argmax(label_, 1)

            # Perform forward pass and compute predictions as the
            # most likely class
            output = model(input_)
            _, predictions = torch.max(output, 1)

            corrects = torch.sum(
                predictions.detach()== int_label_.detach(),
            )

            l = criterion(output, label_)
            loss += float(l.detach().cpu().numpy())

            running_corrects += float(corrects.item())
            running_total += float(label_.size(0))

            all_labels.append(
                int_label_.detach().cpu().numpy()
            )

            all_predictions.append(
                predictions.detach().cpu().numpy()
            )

        norm_loss = loss / len(val_dl)
        acc = running_corrects/running_total
        print('EVAL LOSS: ', norm_loss)
        print('EVAL ACC : ', acc)

    all_predictions = np.concatenate(all_predictions)
    all_labels = np.concatenate(all_labels)
    np.savetxt(
        osp.join(out_path, 'predictions.csv'),
        all_predictions)
    np.savetxt(
        osp.join(out_path, 'labels.csv'),
        all_labels)

    PL = np.stack([all_predictions, all_labels], 0)
    print('Predictions | Labels')
    print(PL.T[:15, :])
    return acc, norm_loss


def train_cv(
    X: Union[np.ndarray, sparse.csr.csr_matrix],
    y: np.ndarray,
    batch_size: int,
    n_epochs: int,
    lr: float,
    optimizer_name: str,    
    weight_decay: float,
    ModelClass: nn.Module,
    fold_indices: list,
    out_path: str,
    n_genes: int = None,
    mixup_alpha: float = None,
    unlabeled_counts: np.ndarray = None,
    unsup_max_weight: float = 2.,
    unsup_mean_teacher: bool = False,
    ssl_method: str='mixmatch',
    ssl_kwargs: dict={},
    weighted_classes: bool = False,
    balanced_classes: bool = False,
    **kwargs,
) -> (np.ndarray, np.ndarray):
    '''Perform training using a provided set of training/hold-out
    sample indices.

    Parameters
    ----------
    X : np.ndarray
        [Cells, Genes] of log1p transformed normalized values.
        log1p and normalization performed using scanpy defaults.
    y : np.ndarray
        [Cells,] integer class labels.
    n_epochs : int
        number of epochs for training.
    weight_decay : float
        weight decay to apply to model weights.
    lr : float
        learning rate.
    optimizer_name : str
        optimizer to use. {"adadelta", "adam"}.        
    ModelClass : nn.Module
        a model class for construction classification models.
    batch_size : int
        batch size for training.        
    fold_indices : list
        elements are 2-tuple, with training indices and held-out.
    out_path : str
        top level path for saving fold outputs.
    n_genes : int
        number of genes in the input. Not necessarily `X.shape[1]` if
        the input matrix has been concatenated with other features.
    mixup_alpha : float
        alpha parameter for an optional MixUp augmentation during training.
    unsup_max_weight : float
        maximum weight for the unsupervised loss term.
    unsup_mean_teacher : bool
        use a mean teacher for pseudolabel generation.
    ssl_method : str
        semi-supervised learning method to use.
    ssl_kwargs : dict
        arguments passed to the semi-supervised learning loss.
    balanced_classes : bool
        perform class balancing by undersampling majority classes.
    weighted_classes : bool
        weight loss for each class based on relative abundance of classes
        in the training data.

    Returns
    -------
    fold_eval_acc : np.ndarray
        evaluation accuracies for each fold.
    fold_eval_losses : np.ndarray
        loss values for each fold.
    '''
    fold_eval_losses = np.zeros(len(fold_indices))
    fold_eval_acc = np.zeros(len(fold_indices))

    # Perform training on each fold specified in `fold_indices`
    for f in range(len(fold_indices)):
        print('Training tissue independent, fold %d.' % f)
        fold_out_path = osp.join(out_path, 'fold' + str(f).zfill(2))

        os.makedirs(fold_out_path, exist_ok=True)

        traintest_idx = fold_indices[f][0].astype('int')
        val_idx = fold_indices[f][1].astype('int')

        acc, loss = fit_model(
            X=X,
            y=y,
            traintest_idx=traintest_idx,
            val_idx=val_idx,
            out_path=fold_out_path,
            batch_size=batch_size,
            n_epochs=n_epochs,
            ModelClass=ModelClass,
            n_genes=n_genes,
            lr=lr,
            optimizer_name=optimizer_name,            
            weight_decay=weight_decay,
            mixup_alpha=mixup_alpha,
            unlabeled_counts=unlabeled_counts,
            unsup_max_weight=unsup_max_weight,
            unsup_mean_teacher=unsup_mean_teacher,
            ssl_method=ssl_method,
            ssl_kwargs=ssl_kwargs,    
            weighted_classes=weighted_classes,
            balanced_classes=balanced_classes,
            **kwargs,
        )
        
        fold_eval_losses[f] = loss
        fold_eval_acc[f] = acc
    return fold_eval_acc, fold_eval_losses


def train_all(
    X: Union[np.ndarray, sparse.csr.csr_matrix],
    y: np.ndarray,
    batch_size: int,
    n_epochs: int,
    ModelClass: nn.Module,
    out_path: str,
    n_genes: int = None,
    lr: float=1.0,
    optimizer_name: str='adadelta',
    weight_decay: float = None,
    mixup_alpha: float = None,
    unlabeled_counts: np.ndarray = None,
    unsup_max_weight: float = 2.,
    unsup_mean_teacher: bool = False,
    ssl_method: str='mixmatch',
    ssl_kwargs: dict={},    
    weighted_classes: bool = False,
    balanced_classes: bool = False,
    **kwargs,
) -> (float, float):
    '''Perform training using all provided samples.

    Parameters
    ----------
    X : np.ndarray
        [Cells, Genes] of log1p transformed normalized values.
        log1p and normalization performed using scanpy defaults.
    y : np.ndarray
        [Cells,] integer class labels.
    n_epochs : int
        number of epochs for training.
    ModelClass : nn.Module
        a model class for construction classification models.
    batch_size : int
        batch size for training.
    out_path : str
        top level path for saving fold outputs.
    n_genes : int
        number of genes in the input. Not necessarily `X.shape[1]` if
        the input matrix has been concatenated with other features.
    lr : float
        learning rate.
    optimizer_name : str
        optimizer to use. {"adadelta", "adam"}.        
    weight_decay : float
        weight decay to apply to model weights.
    balanced_classes : bool
        perform class balancing by undersampling majority classes.
    weighted_classes : bool
        weight loss for each class based on relative abundance of classes
        in the training data.      

    Returns
    -------
    loss : float
        best loss on the testing set used for model selection.
    acc : float
        best accuracy on the testing set used for model selection.
    '''
    # Prepare a unique output directory
    all_out_path = osp.join(out_path, 'all_data')
    if not osp.exists(all_out_path):
        os.mkdir(all_out_path)

    # Generate training and model selection indices
    traintest_idx = np.random.choice(
        np.arange(X.shape[0]),
        size=int(np.floor(0.9*X.shape[0])),
        replace=False,
    ).astype('int')
    val_idx = np.setdiff1d(
        np.arange(X.shape[0]), traintest_idx,
    ).astype('int')

    acc, loss = fit_model(
        X=X,
        y=y,
        traintest_idx=traintest_idx,
        val_idx=val_idx,
        batch_size=batch_size,
        n_epochs=n_epochs,
        ModelClass=ModelClass,
        out_path=all_out_path,
        n_genes=n_genes,
        lr=lr,
        optimizer_name=optimizer_name,        
        weight_decay=weight_decay,
        mixup_alpha=mixup_alpha,
        unlabeled_counts=unlabeled_counts,
        unsup_max_weight=unsup_max_weight,
        unsup_mean_teacher=unsup_mean_teacher,
        ssl_method=ssl_method,
        ssl_kwargs=ssl_kwargs,    
        weighted_classes=weighted_classes,
        balanced_classes=balanced_classes,
        **kwargs,
    )

    np.savetxt(
        osp.join(all_out_path, 'test_loss_acc.csv'),
        np.array([loss, acc]).reshape(2, 1),
        delimiter=',')

    return loss, acc


def train_tissue_independent_cv(
    X: Union[np.ndarray, sparse.csr.csr_matrix],
    metadata: pd.DataFrame,
    out_path: str,
    balanced_classes: bool = False,
    weighted_classes: bool = False,
    batch_size: int = 256,
    n_epochs: int = 200,
    lower_group: str = 'cell_ontology_class',
    **kwargs,
) -> None:
    '''
    Trains a cell type classifier that is independent of tissue origin

    Parameters
    ----------
    X : np.ndarray
        [Cells, Genes] of log1p transformed, normalized values.
        log1p and normalization performed using scanpy defaults.
    metadata : pd.DataFrame
        [Cells, Features] data with `upper_group` and `lower_group` columns.
    out_path : str
        path for saving trained model weights and evaluation performance.
    balanced_classes : bool
        perform class balancing by undersampling majority classes.
    weighted_classes : bool
        weight loss for each class based on relative abundance of classes
        in the training data.
    batch_size : int
        batch size for training.
    n_epochs : int
        number of epochs for training.
    lower_group : str
        column in `metadata` corresponding to output classes. i.e. cell types.

    Returns
    -------
    None.

    Notes
    -----
    Passes `kwargs` to `CellTypeCLF`.
    '''

    print('TRAINING TISSUE INDEPENDENT CLASSIFIER')
    print('-'*20)
    print()

    if not os.path.exists(out_path):
        os.mkdir(out_path)

    # identify all the `lower_group` levels and create
    # an integer class vector corresponding to unique levels
    celltypes = sorted(list(set(metadata[lower_group])))
    print('There are %d %s in the experiment.\n' %
          (len(celltypes), lower_group))

    for t in celltypes:
        print(t)

    # identify all the `lower_group` levels and create
    # an integer class vector corresponding to unique levels
    y = pd.Categorical(metadata[lower_group]).codes
    y = y.astype('int32')
    labels = pd.Categorical(metadata[lower_group]).categories
    # save mapping of levels : integer values as a CSV
    out_df = pd.DataFrame({'label': labels, 'code': np.arange(len(labels))})
    out_df.to_csv(osp.join(out_path, 'celltype_label.csv'))

    # generate k-fold cross-validation split indices
    # & vectors for metrics evaluated at each fold.
    kf = StratifiedKFold(n_splits=5, shuffle=True)
    kf_indices = list(kf.split(X, y))

    # Perform training on each fold specified in `kf_indices`
    fold_eval_acc, fold_eval_losses = train_cv(
        X=X,
        y=y,
        batch_size=batch_size,
        n_epochs=n_epochs,
        ModelClass=CellTypeCLF,
        fold_indices=kf_indices,
        out_path=out_path,
        balanced_classes=balanced_classes,
        weighted_classes=weighted_classes,
        **kwargs,
    )

    # Save the per-fold results to CSVs

    print('Fold eval losses')
    print(fold_eval_losses)
    print('Fold eval accuracy')
    print(fold_eval_acc)
    print('Mean %f Std %f' % (fold_eval_losses.mean(), fold_eval_losses.std()))
    np.savetxt(osp.join(out_path, 'fold_eval_losses.csv',), fold_eval_losses)
    np.savetxt(osp.join(out_path, 'fold_eval_acc.csv',), fold_eval_acc)

    # Train a model using all available data (after class balancing)
    val_loss, val_acc = train_all(
        X=X,
        y=y,
        batch_size=batch_size,
        n_epochs=n_epochs,
        ModelClass=CellTypeCLF,
        out_path=out_path,
        balanced_classes=balanced_classes,
        weighted_classes=weighted_classes,
        **kwargs,
    )

    return


def train_tissue_dependent_cv(
    X: Union[np.ndarray, sparse.csr.csr_matrix],
    metadata: pd.DataFrame,
    out_path: str,
    balanced_classes: bool = False,
    weighted_classes: bool = False,
    batch_size: int = 256,
    n_epochs: int = 200,
    upper_group: str = 'tissue',
    lower_group: str = 'cell_ontology_class',
    **kwargs,
) -> None:
    '''
    Trains a cell type classifier conditioned on tissue of origin.


    Parameters
    ----------
    X : np.ndarray
        [Cells, Genes] of log1p transformed, normalized values.
        log1p and normalization performed using scanpy defaults.
    metadata : pd.DataFrame
        [Cells, Features] data with 'tissue' and 'cell_ontology_class' columns.
    out_path : str
        path for saving trained model weights and evaluation performance.
    balanced_classes : bool, optional
        perform class balancing by undersampling majority classes.
    weighted_classes : bool
        weight loss for each class based on relative abundance of classes
        in the training data.        
    batch_size : int
        batch size for training.
    upper_group : str
        column in `metadata` with subsets for training `lower_group`
        classifiers independently. i.e. tissues.
    lower_group : str
        column in `metadata` corresponding to output classes. i.e. cell types.

    Returns
    -------
    None.

    Notes
    -----
    Appends a one-hot vector indicating the tissue of origin to each input.
    Length of the one-hot vector is equal to `len(set(metadata[upper_group]))`.
    '''

    print('TRAINING TISSUE DEPENDENT CLASSIFIER')
    print('-'*20)
    print()

    if not os.path.exists(out_path):
        os.mkdir(out_path)

    celltypes = sorted(list(set(metadata[lower_group])))
    tissues = sorted(list(set(metadata[upper_group])))

    n_tissues = len(tissues)
    n_genes = X.shape[1]

    print('There are %d %s in the experiment.' % (len(celltypes), lower_group))
    print('There are %d tissues in the experiment.\n' % len(tissues))
    print()
    print('%s levels:' % lower_group)

    for t in celltypes:
        print(t)

    # identify all the `lower_group` levels and create
    # an integer class vector corresponding to unique levels
    y = pd.Categorical(metadata[lower_group]).codes
    y = y.astype('int32')
    labels = pd.Categorical(metadata[lower_group]).categories
    out_df = pd.DataFrame({'label': labels, 'code': np.arange(len(labels))})
    out_df.to_csv(osp.join(out_path, 'celltype_label.csv'))

    # Make `upper_group` one-hot matrix
    tissue_labels = pd.Categorical(metadata[upper_group])
    tissue_idx = np.array(tissue_labels.codes)
    tissue_idx = torch.from_numpy(tissue_idx.astype('int32')).long()
    tissue_categories = np.array(tissue_labels.categories)

    np.savetxt(osp.join(out_path, 'tissue_categories.csv'),
               tissue_categories,
               delimiter=',', fmt='%s')

    one_hot_mat = utils.make_one_hot(tissue_idx, C=len(tissues))
    one_hot_mat = one_hot_mat.numpy()
    assert X.shape[0] == one_hot_mat.shape[0], \
        'dims unequal at %d, %d' % (X.shape[0], one_hot_mat.shape[0])

    # append `upper_group` one hot vector
    # to the [Cells, Genes] matrix.
    print('Input [Cells, Genes] array shape: ', X.shape)
    X = np.concatenate([X, one_hot_mat], axis=1)
    print('One hot matrix appended to [Cells, Genes] array.')
    print('New dimensions: ', X.shape)

    # generate k-fold cross-validation split indices
    # & vectors for metrics evaluated at each fold.
    kf = StratifiedKFold(n_splits=5, shuffle=True)
    kf_indices = list(kf.split(X, y))

    # Perform training on each fold specified in `kf_indices`
    fold_eval_acc, fold_eval_losses = train_cv(
        X=X,
        y=y,
        batch_size=batch_size,
        n_epochs=n_epochs,
        n_genes=n_genes,    # set n_genes and n_tissues for Conditional model
        n_tissues=n_tissues,
        ModelClass=CellTypeCLFConditional,
        fold_indices=kf_indices,
        out_path=out_path,
        balanced_classes=balanced_classes,
        weighted_classes=weighted_classes,
        **kwargs,
    )

    print('Fold eval losses')
    print(fold_eval_losses)
    print('Fold eval accuracy')
    print(fold_eval_acc)
    print('Mean %f Std %f' % (fold_eval_losses.mean(), fold_eval_losses.std()))
    np.savetxt(osp.join(out_path, 'fold_eval_losses.csv',), fold_eval_losses)
    np.savetxt(osp.join(out_path, 'fold_eval_acc.csv',), fold_eval_acc)

    # Train a model using all available data (after class balancing)
    val_loss, val_acc = train_all(
        X=X,
        y=y,
        batch_size=batch_size,
        n_epochs=n_epochs,
        n_genes=n_genes,    # set n_genes and n_tissues for Conditional model
        n_tissues=n_tissues,
        ModelClass=CellTypeCLFConditional,
        out_path=out_path,
        balanced_classes=balanced_classes,
        weighted_classes=weighted_classes,
        **kwargs,
    )
    return


def train_one_tissue_cv(
    X: Union[np.ndarray, sparse.csr.csr_matrix],
    metadata: pd.DataFrame,
    out_path: str,
    balanced_classes: bool = False,
    weighted_classes: bool = False,
    batch_size: int = 256,
    n_epochs: int = 200,
    upper_group: str = 'tissue',
    lower_group: str = 'cell_ontology_class',
    **kwargs,
) -> None:
    '''
    Trains a cell type classifier for a single tissue

    Parameters
    ----------
    X : np.ndarray
        [Cells, Genes] of log1p transformed, normalized values.
        log1p and normalization performed using scanpy defaults.
    metadata : pd.DataFrame
        [Cells, Features] data with `upper_group` and `lower_group` columns.
    out_path : str
        path for saving trained model weights and evaluation performance.
    balanced_classes : bool, optional
        perform class balancing by undersampling majority classes.
    weighted_classes : bool
        weight loss for each class based on relative abundance of classes
        in the training data.        
    upper_group : str
        column in `metadata` with subsets for training `lower_group`
        classifiers independently. i.e. tissues.
    lower_group : str
        column in `metadata` corresponding to output classes. i.e. cell types.

    Returns
    -------
    None.
    '''

    tissue_str = str(list(metadata[upper_group])[0]).lower()
    print('TRAINING %s DEPENDENT CLASSIFIER FOR: ' %
          upper_group.upper(), tissue_str.upper())
    print('-'*20)
    print()

    celltypes = sorted(list(set(metadata[lower_group])))
    print('There are %d %s in the experiment.\n' %
          (len(celltypes), lower_group))
    for t in celltypes:
        print(t)
    print('')
    y = pd.Categorical(metadata[lower_group]).codes
    y = y.astype('int32')
    labels = pd.Categorical(metadata[lower_group]).categories
    out_df = pd.DataFrame({'label': labels, 'code': np.arange(len(labels))})
    out_df.to_csv(osp.join(out_path, 'celltype_label.csv'))

    kf = StratifiedKFold(n_splits=5, shuffle=True)
    kf_indices = list(kf.split(X, y))

    # Perform training on each fold specified in `kf_indices`
    fold_eval_acc, fold_eval_losses = train_cv(
        X=X,
        y=y,
        batch_size=batch_size,
        n_epochs=n_epochs,
        ModelClass=CellTypeCLF,
        fold_indices=kf_indices,
        out_path=out_path,
        weighted_classes=weighted_classes,
        balanced_classes=balanced_classes,
        **kwargs,
    )

    print('Fold eval losses')
    print(fold_eval_losses)
    print('Fold eval accuracy')
    print(fold_eval_acc)
    print('Mean %f Std %f' % (fold_eval_losses.mean(), fold_eval_losses.std()))
    np.savetxt(osp.join(out_path, 'fold_eval_losses.csv',), fold_eval_losses)
    np.savetxt(osp.join(out_path, 'fold_eval_acc.csv',), fold_eval_acc)

    # Train a model using all available data (after class balancing)
    val_loss, val_acc = train_all(
        X=X,
        y=y,
        batch_size=batch_size,
        n_epochs=n_epochs,
        ModelClass=CellTypeCLF,
        out_path=out_path,
        weighted_classes=weighted_classes,
        balanced_classes=balanced_classes,        
        **kwargs,
    )
    return


def train_tissue_specific_cv(
    X: Union[np.ndarray, sparse.csr.csr_matrix],
    metadata: pd.DataFrame,
    out_path: str,
    balanced_classes: dict = None,
    weighted_classes: bool = None,
    upper_group: str = 'tissue',
    lower_group: str = 'cell_ontology_class',
    **kwargs,
) -> None:
    '''
    Train tissue dependent cell type classifiers for each
    tissue in the experiment

    Parameters
    ----------
    X : np.ndarray
        [Cells, Genes] of log1p transformed, normalized values.
        log1p and normalization performed using scanpy defaults.
    metadata : pd.DataFrame
        [Cells, Features] data with 'tissue' and 'cell_ontology_class' columns.
    out_path : str
        path for saving trained model weights and evaluation performance.
    balanced_classes : dict
        keyed by tissue name.
        values are booleans specifying whether to perform class
        balancing for a given tissue.
    weighted_classes : bool
        weight loss for each class based on relative abundance of classes
        in the training data.        
    upper_group : str
        column in `metadata` with subsets for training `lower_group`
        classifiers independently. i.e. tissues.
    lower_group : str
        column in `metadata` corresponding to output classes. i.e. cell types.
    '''
    if not os.path.exists(out_path):
        os.mkdir(out_path)

    # Identify all the `upper_group` levels present in the data
    tissues = sorted(list(set(metadata[upper_group])))
    print('There are %d %s in the experiment.' % (len(tissues), upper_group))
    for t in tissues:
        print(t)

    # For each `upper_group` level, train a classification model
    for i, t in enumerate(tissues):
        tissue_bidx = metadata[upper_group] == t

        tissue_meta = metadata.loc[tissue_bidx, :]
        tissue_X = X[tissue_bidx, :]

        tissue_out_path = osp.join(out_path, str(t))
        if not os.path.exists(tissue_out_path):
            os.mkdir(tissue_out_path)

        if balanced_classes is None:
            bc = False
        else:
            # determine if we should balance classes for this tissue
            bc = balanced_classes.get(t, False)

        train_one_tissue_cv(
            tissue_X,
            tissue_meta,
            tissue_out_path,
            balanced_classes=bc,
            weighted_classes=weighted_classes,
            upper_group=upper_group,
            lower_group=lower_group,
            **kwargs,
        )

    return

#########################################################
# Predict cell types with a trained model
#########################################################


def predict_cell_types(
    X: Union[np.ndarray, sparse.csr.csr_matrix],
    model_path: str,
    out_path: str,
    upper_groups: Union[list, np.ndarray] = None,
    lower_group_labels: list = None,
    **kwargs,
) -> None:
    '''Predict cell types using a pretrained model

    Parameters
    ----------
    X : np.ndarray, sparse.csr.csr_matrix
        [Cells, Genes] of log1p transformed, normalized values.
        log1p and normalization performed using scanpy defaults.
    model_path : str
        path to a set of pretrained model weights.
    out_path : str
        path for prediction outputs.
    upper_groups : list, np.ndarray
        [Cells,] iterable of str specifying the `upper_group` for each cell.
        if provided, assumes an `upper_group` conditional model.
        if `None`, assumes an `upper_group` independent model.
    lower_group_labels : list
        str labels corresponding to output nodes of the model.

    Returns
    -------
    None.

    Notes
    -----
    `**kwargs` passed to `scnym.predict.Predicter`.
    '''
    if upper_groups is not None:
        print('Assuming conditional model.')

        X, categories = utils.append_categorical_to_data(X, upper_groups)
        np.savetxt(osp.join(out_path, 'category_names.csv'),
                   categories,
                   fmt='%s',
                   delimiter=',')
    else:
        print('Assuming independent model')

    # Intantiate a prediction object, which handles batch processing
    P = Predicter(
        model_weights=model_path,
        n_genes=X.shape[1],
        n_cell_types=None,  # infer cell type # from weights
        labels=lower_group_labels,
        **kwargs,
    )

    predictions, names, scores = P.predict(X, output='score')

    probabilities = F.softmax(torch.from_numpy(scores), dim=1)
    probabilities = probabilities.cpu().numpy()

    np.savetxt(osp.join(out_path, 'predictions_idx.csv'),
               predictions, delimiter=',')
    np.savetxt(osp.join(out_path, 'probabilities.csv'),
               probabilities, delimiter=',')
    np.savetxt(osp.join(out_path, 'raw_scores.csv'),
               scores, delimiter=',')
    if names is not None:
        np.savetxt(osp.join(out_path, 'predictions_names.csv'),
                   names, delimiter=',', fmt='%s')
    return


#########################################################
# utilities
#########################################################


def load_data(
    path: str,
) -> Union[np.ndarray, sparse.csr.csr_matrix]:
    '''Load a counts matrix from a file path.

    Parameters
    ----------
    path : str
        path to [npy, csv, h5ad, loom] file.

    Returns
    -------
    X : np.ndarray
        [Cells, Genes] matrix.
    '''
    if osp.splitext(path)[-1] == '.npy':
        print('Assuming sparse matrix...')
        X_raw = np.load(path, allow_pickle=True)
        X_raw = X_raw.item()
    elif osp.splitext(path)[-1] == '.csv':
        X_raw = np.loadtxt(path, delimiter=',')
    elif osp.splitext(path)[-1] == '.h5ad':
        adata = sc.read_h5ad(path)
        X_raw = utils.get_adata_asarray(adata=adata)
    elif osp.splitext(path)[-1] == '.loom':
        adata = sc.read_loom(path)
        X_raw = utils.get_adata_asarray(adata=adata)
    else:
        raise ValueError('unrecognized file type %s for counts' %
                         osp.splitext(path)[-1])

    return X_raw

#########################################################
# main()
#########################################################


def main():
    import configargparse
    import yaml
    parser = configargparse.ArgParser(
        description='Train cell type classifiers',
        default_config_files=['./configs/default_config.txt'])
    parser.add_argument('command', type=str,
                        help='action to perform. \
                       ["train_tissue_independent", \
                       "train_tissue_dependent", \
                       "train_tissue_specific", \
                       "find_cell_type_markers", \
                       "predict_cell_types"]')
    parser.add_argument('-c', is_config_file=True, required=False,
                        help='path to a configuration file.')
    parser.add_argument('--input_counts', type=str, required=True,
                        help='path to input data [Cells, Genes] counts. \
                        [npy, csv, h5ad, loom]')
    parser.add_argument('--input_gene_names', type=str, required=True,
                        help='path to gene names for the input data.')
    parser.add_argument('--training_gene_names', type=str, required=False,
                        help='path to training data gene names. \
                        required for prediction.')
    parser.add_argument('--training_metadata', type=str, required=True,
                        help='CSV metadata for training. Requires `upper_group` and `lower_group` columns. \
                       necessary for prediction to provide cell type names.')
    parser.add_argument('--lower_group', type=str, required=True,
                        default='cell_ontology_class',
                        help='column in `metadata` with to output labels. \
                        i.e. cell types.')
    parser.add_argument('--upper_group', type=str, required=True,
                        default='tissue',
                        help='column in `metadata` with to subsets for independent training. \
                        i.e. tissues.')
    parser.add_argument('--out_path', type=str, required=True,
                        help='path for output files')
    parser.add_argument('--genes_to_use', type=str, default=None,
                        help='path to a text file of genes to use for training. \
                       must be a subset of genes in `training_gene_names`')
    parser.add_argument(
        '--input_domain_group',
        type=str,
        help='column in `training_metadata` that specifies domain of origin for each training observation.',
        required=False,
        default=None,
    )
    parser.add_argument('--batch_size', type=int, default=256,
                        help='batch size for training')
    parser.add_argument('--n_epochs', type=int, default=256,
                        help='number of epochs for training')
    parser.add_argument('--init_dropout', type=float, default=0.3,
                        help='initial dropout to perform on gene inputs')
    parser.add_argument('--n_hidden', type=int, default=128,
                        help='number of hidden units in the classifier')
    parser.add_argument('--n_layers', type=int, default=2,
                        help='number of hidden layers in the model')
    parser.add_argument('--residual', action='store_true',
                        help='use residual layers in the model')
    parser.add_argument('--track_running_stats', type=bool, default=True,
                       help='track running statistics in batch normalization layers')
    parser.add_argument('--model_path', type=str, default=None,
                        help='path to pretrained model weights \
                        for class marker identification.')
    parser.add_argument('--weight_decay', type=float, default=1e-5,
                        help='weight decay applied by the optimizer')
    parser.add_argument('--lr', type=float, default=1.0,
                        help='learning rate for the optimizer.')
    parser.add_argument('--optimizer', type=str, default='adadelta',
                       help='optimizer to use. {adadelta, adam}.')
    parser.add_argument('--l1_reg', type=float, default=1e-4,
                        help='l1 regularization strength \
                        for class marker identification')
    parser.add_argument('--weight_classes', type=bool, default=False,
                        help='weight loss based on relative class abundance.')
    parser.add_argument('--balance_classes', type=bool, default=False,
                        help='perform class balancing.')
    parser.add_argument('--mixup_alpha', type=float, default=None,
                        help='alpha parameter for MixUp training. \
                        if set performs MixUp, otherwise does not.')
    parser.add_argument('--unlabeled_counts', type=str, default=None,
                        help='path to unlabeled data [Cells, Genes]. \
                       [npy, csv, h5ad, loom]. \
                       if provided, uses interpolation consistency training.')
    parser.add_argument('--unlabeled_genes', type=str, default=None,
                        help='path to gene names for the unlabeled data.\
                       if not provided, assumes same as `input_counts`.')
    parser.add_argument(
        '--unlabeled_domain',
        type=str,
        help='path to a CSV of integer domain labels for each data point in `unlabeled_counts`.',
        required=False,
        default=None,
    )
    parser.add_argument('--unsup_max_weight', type=float, default=2.,
                        help='maximum weight for the unsupervised component of IC training.')
    parser.add_argument('--unsup_mean_teacher', action='store_true',
                        help='use a mean teacher for IC training.')
    parser.add_argument(
        '--ssl_method',
        type=str,
        default='mixmatch',
        help='semi-supervised learning method to use. {"mixmatch", "ict"}.',
    )
    parser.add_argument(
        '--ssl_config',
        type=str,
        default=None,
        help='path to a YAML configuration file of kwargs for the SSL method.'
    )
    args = parser.parse_args()

    print(args)
    print(parser.format_values())

    COMMANDS = [
        'train_tissue_independent',
        'train_tissue_dependent',
        'train_tissue_specific',
        'predict_cell_types',
    ]

    if args.command not in COMMANDS:
        raise ValueError('%s is not a valid command.' % args.command)

    #####################################
    # LOAD DATA
    #####################################

    X_raw = load_data(args.input_counts)

    print('Loaded data.')
    print('%d cells and %d genes in raw data.' % X_raw.shape)
    gene_names = np.loadtxt(args.input_gene_names, dtype='str')
    print('Loaded gene names for the raw data. %d genes.' % len(gene_names))

    if args.genes_to_use is not None:
        genes_to_use = np.loadtxt(args.genes_to_use, dtype='str')
        print('Using a subset of %d genes as specified in \n %s.' %
              (len(genes_to_use), args.genes_to_use))
    else:
        genes_to_use = gene_names

    if args.genes_to_use is not None:
        # Filter the input matrix to use only the specified genes
        print('Using %d genes for classification.' % len(genes_to_use))
        gnl = gene_names.tolist()
        keep_idx = np.array([gnl.index(x) for x in genes_to_use])
        X = X_raw[:, keep_idx]
    else:
        # leave all genes in the matrix
        X = X_raw

    # Load metadata and identify output classes
    metadata = pd.read_csv(args.training_metadata,)
    lower_groups = np.unique(metadata[args.lower_group]).tolist()
    
    # load domain labels if applicable
    if args.input_domain_group is not None:
        if args.input_domain_group not in metadata.columns:
            msg = f'{args.input_domain_group} is not a column in `training_metadata`'
            raise ValueError(msg)
        else:
            input_domain = np.array(metadata[args.input_domain_group])
    else:
        input_domain = None

    # Load any provided unlabeled data for semi-supervised learning
    if args.unlabeled_counts is not None:
        unlabeled_counts = load_data(args.unlabeled_counts)
        print('%d cells, %d genes in unlabeled data.' % unlabeled_counts.shape)
        
        # parse any semi-supervised learning specific parameters
        if args.ssl_config is not None:
            print(f'Loading Semi-Supervised Learning parameters for {args.ssl_method}')
            with open(args.ssl_config, 'r') as f:
                ssl_kwargs = yaml.load(f, Loader=yaml.Loader)
            print('SSL kwargs:')
            for k, v in ssl_kwargs.items():
                print(f'{k}\t\t:\t\t{v}')
            print()
        else:
            ssl_kwargs = {}
        
    else:
        unlabeled_counts = None
        ssl_kwargs = {}

    if args.unlabeled_genes is not None and unlabeled_counts is not None:
        # Contruct a matrix using the unlabeled counts where columns
        # correspond to the same gene in `input_counts`.
        print('Subsetting unlabeled counts to genes used for training...')
        unlabeled_genes = np.loadtxt(
            args.unlabeled_genes,
            delimiter=',',
            dtype='str',
        )
        unlabeled_counts = utils.build_classification_matrix(
            X=unlabeled_counts,
            model_genes=genes_to_use,
            sample_genes=unlabeled_genes,
        )
        if args.unlabeled_domain is not None:
            unlabeled_domain = np.loadtxt(
                args.unlabeled_domain,
            ).astype(np.int)
        else:
            unlabeled_domain = None
    else:
        unlabeled_domain = None

    # prepare output paths
    if not os.path.exists(args.out_path):
        os.mkdir(args.out_path)

    sub_dirs = [
        'tissues', 
        'tissue_independent_no_dropout',
        'tissue_dependent', 
        'tissue_ind_class_optimums',
    ]
    for sd in sub_dirs:
        if not os.path.exists(osp.join(args.out_path, sd)):
            os.mkdir(osp.join(args.out_path, sd))

    #####################################
    # TISSUE INDEPENDENT CLASSIFIERS
    #####################################

    if args.command == 'train_tissue_independent':
        train_tissue_independent_cv(
            X,
            metadata,
            osp.join(args.out_path, 'tissue_independent'),
            balanced_classes=args.balance_classes,
            weighted_classes=args.weight_classes,
            batch_size=args.batch_size,
            n_epochs=args.n_epochs,
            init_dropout=args.init_dropout,
            lower_group=args.lower_group,
            n_hidden=args.n_hidden,
            n_layers=args.n_layers,
            lr=args.lr,
            optimizer_name=args.optimizer,
            weight_decay=args.weight_decay,
            residual=args.residual,
            track_running_stats=args.track_running_stats,
            mixup_alpha=args.mixup_alpha,
            unlabeled_counts=unlabeled_counts,
            unsup_max_weight=args.unsup_max_weight,
            unsup_mean_teacher=args.unsup_mean_teacher,
            ssl_method=args.ssl_method,
            ssl_kwargs=ssl_kwargs,
            input_domain=input_domain,
            unlabeled_domain=unlabeled_domain,
        )

    #####################################
    # TISSUE DEPENDENT CLASSIFIERS
    #####################################

    if args.command == 'train_tissue_dependent':
        train_tissue_dependent_cv(
            X,
            metadata,
            osp.join(args.out_path, 'tissue_dependent'),
            balanced_classes=args.balance_classes,
            weighted_classes=args.weight_classes,
            batch_size=args.batch_size,
            n_epochs=args.n_epochs,
            upper_group=args.upper_group,
            lower_group=args.lower_group,
            n_hidden=args.n_hidden,
            n_layers=args.n_layers,
            lr=args.lr,
            optimizer_name=args.optimizer,
            weight_decay=args.weight_decay,
            residual=args.residual,
            track_running_stats=args.track_running_stats,
            mixup_alpha=args.mixup_alpha,
            unlabeled_counts=unlabeled_counts,
            unsup_max_weight=args.unsup_max_weight,
            unsup_mean_teacher=args.unsup_mean_teacher,
            ssl_method=args.ssl_method,
            ssl_kwargs=ssl_kwargs,
            input_domain=input_domain,
            unlabeled_domain=unlabeled_domain,
        )

    #####################################
    # TISSUE SPECIFIC CLASSIFIERS
    #####################################

    if args.command == 'train_tissue_specific':

        tissue_specific_bal = {
            k: False for k in set(metadata[args.upper_group])}
        if args.balance_classes:
            print(args.balance_classes)
            for k in tissue_specific_bal:
                tissue_specific_bal[k] = True
            print('Balancing classes on a tissue specific basis')
            print(tissue_specific_bal)

        train_tissue_specific_cv(
            X,
            metadata,
            osp.join(args.out_path, 'tissues'),
            balanced_classes=tissue_specific_bal,
            weighted_classes=args.weight_classes,
            batch_size=args.batch_size,
            n_epochs=args.n_epochs,
            upper_group=args.upper_group,
            lower_group=args.lower_group,
            n_hidden=args.n_hidden,
            n_layers=args.n_layers,
            lr=args.lr,
            optimizer_name=args.optimizer,
            weight_decay=args.weight_decay,
            residual=args.residual,
            track_running_stats=args.track_running_stats,
            unlabeled_counts=unlabeled_counts,
            unsup_max_weight=args.unsup_max_weight,
            unsup_mean_teacher=args.unsup_mean_teacher,
            ssl_method=args.ssl_method,
            input_domain=input_domain,
            unlabeled_domain=unlabeled_domain,
         )

    #####################################
    # PRETRAINED MODEL PREDICTION
    #####################################

    if args.command == 'predict_cell_types':
        if args.model_path is None:
            raise ValueError('`model_path` required.')
        if args.training_gene_names is None:
            raise ValueError('must supply `training_gene_names`.')
        training_genes = np.loadtxt(
            args.training_gene_names, delimiter=',', dtype='str').tolist()

        X = utils.build_classification_matrix(
            X=X,
            model_genes=training_genes,
            sample_genes=gene_names,
        )

        predict_cell_types(
            X,
            model_path=args.model_path,
            out_path=args.out_path,
            lower_group_labels=lower_groups,
            n_hidden=args.n_hidden,
            n_layers=args.n_layers,
            residual=args.residual,
        )


#########################################################
# __main__
#########################################################


if __name__ == '__main__':

    main()



'''Baseline classification models used for comparison to scNym

This module contains all baseline models implemented in python.
'''
import numpy as np
import pandas as pd
import anndata
from sklearn.metrics import pairwise_distances
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import LinearSVC
from sklearn.calibration import CalibratedClassifierCV # platt scaling for SVMs
from scipy import stats
# scanpy for `harmony_clf`
import scanpy as sc
import scanpy.external as sce
import scvi # note -- requires `scvi-tools`
import pickle


############################################
# scmap-cell-exact classification scoring
############################################


def scmap_cell_classif(
    train_adata: anndata.AnnData, 
    test_adata: anndata.AnnData, 
    cell_type_col: str,
    n_neighbors: int=10,
    n_features: int=1000,
    return_prob: bool=False,
    **kwargs,
) -> np.ndarray:
    '''Train a kNN classifier with cosine distances per scmap-cell.
    Instead of an approximate kNN search, this version uses the exact
    kNN search provided in `scikit-learn`.
    
    Parameters
    ----------
    train_adata : anndata.AnnData
        [Cells, Genes] for training.
    test_adata : anndata.AnnData
        [Cells, Genes] for testing
    cell_type_col : str
        column in `.obs` containing cell types for `adata` objects.
    n_neighbors : int
        number of nearest neighbors to use.
    n_features : int
        features to select with the `M3Drop` scheme.
        authors recommend 1000.
    return_prob : bool
        return probabilities in addition to predictions, in that
        order.
    
    Returns
    -------
    predictions : np.ndarray
        [Cells,] predicted labels.
    probabilities : pd.DataFrame, optional
        [Cells, Classes] probability matrix. classes are column names.
        
    Notes
    -----
    The top `n_features` genes are selected using the M3Drop method.
    Default values for `n_features` and `n_neighbors` are set based on the
    hyperparameter optimization presented in the original `scmap` paper.
    
    References
    ----------
    scmap: projection of single-cell RNA-seq data across data sets
    Vladimir Yu Kiselev, Andrew Yiu & Martin Hemberg
    https://www.nature.com/articles/nmeth.4644
    '''
    knn = KNeighborsClassifier(
        n_neighbors=n_neighbors,
        metric='cosine',
        n_jobs=-1,
    )
    
    if n_features is None:
        n_features = train_adata.shape[1]
    
    if n_features < train_adata.shape[1]:
        # fit "M3Drop" model to estimate genes with higher than expected dropout
        if type(train_adata.X) != np.ndarray:
            X = train_adata.X.toarray()
        expr = np.expm1(X).mean(0)
        # percentage of dropouts for each gene
        drop = (1 - (X>0).mean(0)) * 100

        expr = np.log2(expr + 1)
        drop = np.log2(drop)

        lm = stats.linregress(expr, drop)
        pred = lm.slope*expr + lm.intercept
        res = drop - pred
        selection_idx = np.argsort(res)[-n_features:]  
        selected_genes = train_adata.var_names[selection_idx]
    else:
        selected_genes = train_adata.var_names
    
    # remove genes not expressed in the test data
    # per the scmap-cell paper
    selected_genes = np.intersect1d(
        np.array(test_adata.var_names),
        selected_genes,
    )
    
    train_adata = train_adata[:, selected_genes].copy()
    test_adata = test_adata[:, selected_genes].copy()
    
    knn.fit(train_adata.X, train_adata.obs[cell_type_col])
    predictions = knn.predict(test_adata.X)
    probabilities = knn.predict_proba(test_adata.X)
    
    probabilities = pd.DataFrame(
        probabilities,
        columns=knn.classes_,
    )
    
    if return_prob:
        r = (predictions, probabilities)
    else:
        r = predictions
    return r


def vanilla_svm(
    train_adata: anndata.AnnData, 
    test_adata: anndata.AnnData, 
    cell_type_col: str,
    return_prob: bool=False,
    **kwargs,
) -> np.ndarray:
    '''Train and predict with a linear support vector machine.
    
    Parameters
    ----------
    train_adata : anndata.AnnData
        [Cells, Genes] for training.
    test_adata : anndata.AnnData
        [Cells, Genes] for testing.
    cell_type_col : str
        column in `.obs` containing cell types for `adata` objects.        
    return_prob : bool
        return probabilities in addition to predictions, in that
        order. SVMs do not have native probablistic predictions,
        so we scale the decision confidence for a pseudoprobability.

    Returns
    -------
    predictions : np.ndarray
        [Cells,] predicted labels for test set.
        
    Notes
    -----
    A linear support vector machine (SVM) was shown to be the top performing
    baseline in a benchmarking study (see References).
    We set the default SVM hyperparameter values based on this 
    benchmarking paper.
    
    SVMs do not offer probabilistic outputs natively. To approximate a probablistic
    output, we use the Platt Scaling method.
    
    References
    ----------
    A comparison of automatic cell identification methods for \
    single-cell RNA sequencing data.
    Abdelaal, T. et al. 
    Genome Biology 20, 194 (2019).
    
    Probabilistic Outputs for Support Vector Machines and Comparisons \
    to Regularized Likelihood Methods. 
    Platt, J. C.
    Advances in Large Margin Classifiers 61–74 (MIT Press, 1999).
    '''
    model = LinearSVC(**kwargs)
    model.fit(
        train_adata.X,
        train_adata.obs[cell_type_col],
    )
    
    predictions = model.predict(
        test_adata.X,
    )
    
    # SVMs do not have native probabilistic prediction
    # Rather, we scale the model confidence scores 
    # derived from the distance to the support vectors
    # to generate a pseudoprobability with Platt scaling
    #
    # See the relevant `sklearn` documentation
    # https://bit.ly/2FJX21r
    
    # use CalibratedClassifierCV to perform platt scaling
    calib = CalibratedClassifierCV(
        model,
        method='sigmoid', # corresponds to Platt Scaling
        cv=5,
    )
    calib.fit(
        train_adata.X, 
        train_adata.obs[cell_type_col],
    )
    
    # use decision_function
    probabilities = calib.predict_proba(test_adata.X)
    
    probabilities = pd.DataFrame(
        probabilities,
        columns=calib.classes_,
    )
    
    if return_prob:
        r = (predictions, probabilities)
    else:
        r = predictions
    
    return r


def harmony_clf(
    train_adata: anndata.AnnData, 
    test_adata: anndata.AnnData, 
    cell_type_col: str,
    return_prob: bool=False,
    **kwargs,    
) -> (np.ndarray, anndata.AnnData):
    '''Perform `harmony` data integration, followed by SVM classification.
    
    Parameters
    ----------
    train_adata : anndata.AnnData
        [Cells, Genes] for training.
    test_adata : anndata.AnnData
        [Cells, Genes] for testing.
    cell_type_col : str
        column in `.obs` containing cell types for `adata` objects.        
    return_prob : bool
        return probabilities in addition to predictions, in that
        order. SVMs do not have native probablistic predictions,
        so we scale the decision confidence for a pseudoprobability.        
        
    Returns
    -------
    predictions : np.ndarray
        [Cells,] predicted labels for test set.
    probabiliteis : np.ndarray, optional.
        [Cells, Classes]
    joint : anndata.AnnData
        [Cells, Genes] of joined training and test data.
        `.obs['batch']` specifies train (`'0'`) or test (`'1'`) data.
        `.obsm['X_pca_harmony']` contains the harmonized embedding.
        
    Notes
    -----
    Data are first integrated using the `harmonypy` implementation of 
    the Harmony algorithm. An SVM is then trained on the integrated latent
    variables.
    
    See Also
    --------
    vanilla_svm
    '''
    # `.concatenate` automatically adds a batch variable
    joint = train_adata.concatenate(test_adata)
    
    # fit unharmonized PCA
    sc.pp.highly_variable_genes(joint, n_top_genes=4000)
    sc.pp.pca(joint)
    
    # creates "harmonized" PCA embedding
    # `joint.obsm['X_pca_harmony']`
    sce.pp.harmony_integrate(
        joint,
        key='batch',
    )
    
    # train an SVM
    X_train = joint.obsm['X_pca_harmony'][joint.obs['batch']=='0']
    y_train = np.array(joint.obs[cell_type_col])[joint.obs['batch']=='0']
    X_test  = joint.obsm['X_pca_harmony'][joint.obs['batch']=='1']
    y_test  = np.array(joint.obs[cell_type_col])[joint.obs['batch']=='1']
    
    model = LinearSVC(**kwargs)
    model.fit(
        X_train,
        y_train,
    )
    
    predictions = model.predict(
        X_test,
    )
    
    # use CalibratedClassifierCV to perform platt scaling
    calib = CalibratedClassifierCV(
        model,
        method='sigmoid', # corresponds to Platt Scaling
        cv=5,
    )
    calib.fit(
        train_adata.X, 
        train_adata.obs[cell_type_col],
    )
    
    # use decision_function
    probabilities = calib.predict_proba(test_adata.X)
    
    probabilities = pd.DataFrame(
        probabilities,
        columns=calib.classes_,
    )
    
    if return_prob:
        r = (predictions, probabilities, joint)
    else:
        r = (predictions, joint)
    
    return r


def isinteger(x) -> bool:
    '''Check if values in `x` are integers, regardless of
    dtype. Returns elementwise `bool`.
    '''
    # check if X is sparse
    if type(x) == np.ndarray:
        return np.equal(np.mod(x, 1), 0)
    else:
        return np.all(np.equal(np.mod(x.data, 1), 0))


def scanvi(
    train_adata: anndata.AnnData,
    test_adata: anndata.AnnData,
    cell_type_col: str,
    n_per_class: int=100,
    **kwargs,
) -> (np.ndarray, pd.DataFrame, anndata.AnnData, scvi.model.scanvi.SCANVI):
    '''Use SCANVI to transfer annotations.
    
    Parameters
    ----------
    train_adata : anndata.AnnData
        [Cells, Genes] for training.
    test_adata : anndata.AnnData
        [Cells, Genes] for testing.
    cell_type_col : str
        column labeling ground truth cell types in
        `train_adata` and `test_adata`.
    n_per_class : int
        number of training examples per class. scANVI authors
        recommend `100` (default).
    
    Returns
    -------
    predictions : np.ndarray
        [Cells,] cell type label predictions.
    probabilities : pd.DataFrame
        [Cells, Class] probabilities.
        classes are column labels.
    adata : anndata.AnnData
        [Cells, Genes] concatenation of `train_adata` and `test_adata`
        with the learned scANVI latent space in `.obsm['X_scANVI']`.
        class predictions are in `.obs['C_scANVI']`.
    lvae : scvi.model.scanvi.SCANVI
        a trained scANVI model object.
    
    Notes
    -----
    This implementation exactly follows the working example
    from the `scvi` authors.
    
    https://www.scvi-tools.org/en/0.7.0-alpha.4/user_guide/notebooks/harmonization.html
    '''
    # check that train_adata and test_adata contain
    # raw counts
    tr_int = isinteger(train_adata.X)
    te_int = isinteger(test_adata.X)
    if not (tr_int and te_int):
        # check if the raw counts are in `.raw`
        tr_int = isinteger(train_adata.raw.X)
        te_int = isinteger(test_adata.raw.X)
        if tr_int and te_int:
            train_adata = train_adata.copy()
            test_adata = test_adata.copy()
            # set raw counts to `X`
            train_adata.X = train_adata.raw[:, train_adata.var_names].X
            test_adata.X = test_adata.raw[:, test_adata.var_names].X
        else:
            msg = 'Integer raw counts not found.'
            raise ValueError(msg)
    else:
        # raw counts are already set to X
        pass
    
    # `.concatenate()` creates batch labels in `.obs['batch']`
    adata = train_adata.concatenate(test_adata)
    
    # store raw counts in a new layer
    # normalize and select highly variable genes
    #
    # scVI uses only a set of highly variable genes
    # to perform data integration in the latent space
    adata.layers["counts"] = adata.X.copy()
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)
    adata.raw = adata  # keep full dimension safe
    sc.pp.highly_variable_genes(
        adata, 
        flavor="seurat_v3", 
        n_top_genes=2000, 
        layer="counts", 
        batch_key="batch",
        subset=True
    )
    
    # assign cell type labels for training
    # scANVI uses a special token `"Unknown"` for cells that are not labeled
    # the authors recommend using 100 cells per cell type
    # from the training set to balance classes
    adata.obs[cell_type_col] = pd.Categorical(
        adata.obs[cell_type_col],
    )
    
    labels = np.repeat("Unknown", adata.shape[0])
    labels = labels.astype("<U43")
    for x in np.unique(adata.obs[cell_type_col]):
        idx = np.where((adata.obs[cell_type_col] == x) & (adata.obs["batch"] == "0"))[
            0
        ]
        sampled = np.random.choice(
            idx, 
            np.min([n_per_class, len(idx)])
        )
        labels[sampled] = adata.obs[cell_type_col][sampled]

    adata.obs["celltype_scanvi"] = labels
    
    # setup scANVI for training
    scvi.data.setup_anndata(
        adata, 
        layer="counts",
        batch_key="batch", 
        labels_key="celltype_scanvi", 
    )
    
    # fit the semi-supervised scANVI model
    lvae = scvi.model.SCANVI(
        adata,
        "Unknown",
        use_cuda=True,
        n_latent=30,
        n_layers=2,
    )
    
    lvae.train(n_epochs_semisupervised=100)
    
    # extract labels
    adata.obs["C_scANVI"] = lvae.predict(adata)
    adata.obsm["X_scANVI"] = lvae.get_latent_representation(adata)
    
    predictions = np.array(adata.obs.loc[adata.obs['batch']=='1', "C_scANVI"])
    # returns a [Cells, Classes] data frame with class
    # names as column labels and cell barcodes as indices
    probabilities = lvae.predict(adata, soft=True)
    probabilities = probabilities.loc[adata.obs['batch']=='1']
    # scANVI will add the "Unknown" token as a class, usually
    # with very low probability
    # here we drop it, then renorm probabilities to == 1
    probabilities = probabilities.drop(columns=['Unknown'])
    probabilities = probabilities / np.tile(
        np.array(probabilities.sum(1)).reshape(-1, 1), (1, probabilities.shape[1])
    )
    # check probability normalization
    eq1 = np.allclose(
        probabilities.sum(1),
        np.ones(probabilities.shape[0]),
    )
    if not eq1:
        msg = 'Not all sum(probabilities) are close to 1.'
        n = np.sum(probabilities.sum(1)!=1.)
        msg += f'{n} cells have probabilities != 1.'
        raise ValueError(msg)

    
    r = (
        predictions, 
        probabilities, 
        adata, 
        lvae,
    )
    
    return r



import torch
import numpy as np
from scipy import sparse
from torch.utils.data import Dataset
from typing import Callable, Any, Union


class SingleCellDS(Dataset):
    '''Dataset class for loading single cell profiles.

    Attributes
    ----------
    X : np.ndarray, sparse.csr_matrix
        [Cells, Genes] cell profiles.
    y_labels : np.ndarray, sparse.csr_matrix
        [Cells,] integer class labels.
    y : torch.FloatTensor
        [Cells, Classes] one hot labels.
    transform : Callable
        performs data transformation operations on a
        `sample` dict.
    num_classes : int
        number of classes in the dataset. default `-1` infers
        the number of classes as `len(unique(y))`.
    '''

    def __init__(
        self,
        X: Union[sparse.csr.csr_matrix, np.ndarray],
        y: Union[sparse.csr.csr_matrix, np.ndarray],
        domain: Union[sparse.csr.csr_matrix, np.ndarray] = None,
        transform: Callable = None,
        num_classes: int=-1,
        num_domains: int=-1,
    ) -> None:
        '''
        Load single cell expression profiles.

        Parameters
        ----------
        X : np.ndarray, sparse.csr_matrix
            [Cells, Genes] expression count matrix.
            scNym tools expect ln(Counts Per Million + 1).
        y : np.ndarray, sparse.csr_matrix
            [Cells,] integer cell type labels.
        domain : np.ndarray, sparse.csr_matrix
            [Cells,] integer domain labels.
        transform : Callable
            transform to apply to samples.
        num_classes : int
            total number of classes for the task.
        num_domains : int
            total number of domains for the task.

        Returns
        -------
        None.
        '''
        super(SingleCellDS, self).__init__()
        
        # check types on input arrays
        if type(X) not in (np.ndarray, sparse.csr_matrix,):
            msg = f'X is type {type(X)}, must `np.ndarray` or `sparse.csr_matrix`'
            raise TypeError(msg)

        if type(y) not in (np.ndarray, sparse.csr_matrix,):
            msg = f'X is type {type(y)}, must `np.ndarray` or `sparse.csr_matrix`'
            raise TypeError(msg)
            
        if type(y) != np.ndarray:
            # densify labels
            y = y.toarray()
        
        self.X = X
        self.y_labels = torch.from_numpy(y).long()
        self.y = torch.nn.functional.one_hot(
            self.y_labels,
            num_classes=num_classes,
        ).float()
        
        self.dom_labels = domain
        if self.dom_labels is not None:
            self.dom = torch.nn.functional.one_hot(
                torch.from_numpy(self.dom_labels).long(),
                num_classes=num_domains,
            ).float()
        else:
            self.dom = np.zeros_like(self.y) - 1
        
        self.transform = transform

        if not self.X.shape[0] == self.y.shape[0]:
            sizes = (self.X.shape[0], self.y.shape[0])
            raise ValueError(
                'X rows %d not equal to y rows %d.' % sizes)
        return

    def __len__(self,) -> int:
        '''Return the number of examples in the data set.'''
        return self.X.shape[0]

    def __getitem__(self, idx: int,) -> dict:
        '''Get a single cell expression profile and corresponding label.

        Parameters
        ----------
        idx : int
            index value in `range(len(self))`.

        Returns
        -------
        sample : dict
            'input' - torch.FloatTensor, input vector
            'output' - torch.LongTensor, target label
        '''
        if type(idx) != int:
            raise TypeError(
                f'indices must be int, you passed {type(idx)}, {idx}'
            )
        
        # check if the idx value is valid given the dataset size
        if idx < 0 or idx > len(self):
            vals = (idx, len(self))
            raise ValueError(
                'idx %d is invalid for dataset with %d examples.' % vals)

        # retrieve relevant sample vector and associated label
        # store in a hash table for later manipulation and retrieval
        
        # input_ is either an `np.ndarray` or `sparse.csr.csr_matrix`
        input_ = self.X[idx, ...]
        # label is already a `torch.Tensor`
        label = self.y[idx]
        
        # if the corresponding vectors are sparse, convert them to dense
        # we perform this operation on a samplewise-basis to avoid
        # storing the whole count matrix in dense format
        if type(input_) != np.ndarray:
            input_ = input_.toarray()
            
        input_ = torch.from_numpy(input_).float()
        if input_.size(0) == 1:
            input_ = input_.squeeze()
        
        sample = {
            'input': input_,
            'output': label,
        }
        
        sample['domain'] = self.dom[idx]

        # if a transformer was supplied, apply transformations
        # to the sample vector and label
        if self.transform is not None:
            sample = self.transform(sample)
        return sample


def balance_classes(
    y: np.ndarray,
    class_min: int = 256,
) -> np.ndarray:
    '''
    Perform class balancing by undersampling majority classes
    and oversampling minority classes, down to a minimum value.

    Parameters
    ----------
    y : np.ndarray
        class assignment indices.
    class_min : int
        minimum number of examples to use for a class.
        below this value, minority classes will be oversampled
        with replacement.

    Returns
    -------
    all_idx : np.ndarray
        indices for balanced classes. some indices may be repeated.
    '''
    # determine the size of the smallest class
    # if < `class_min`, we oversample to `class_min` samples.
    classes, counts = np.unique(y, return_counts=True)
    min_count = int(np.min(counts))
    if min_count < class_min:
        min_count = class_min

    # generate indices with equal representation of each class
    all_idx = []
    for i, c in enumerate(classes):
        class_idx = np.where(y == c)[0].astype('int')
        rep = counts[i] < min_count  # oversample minority classes
        if rep:
            print('Count for class %s is %d. Oversampling.' % (c, counts[i]))
        ridx = np.random.choice(class_idx, size=min_count, replace=rep)
        all_idx += [ridx]
    all_idx = np.concatenate(all_idx).astype('int')
    return all_idx


class LibrarySizeNormalize(object):
    '''Perform library size normalization.'''

    def __init__(
        self, 
        counts_per_cell_after: int = int(1e6),
        log1p: bool=True,
    ) -> None:
        self.counts_per_cell_after = counts_per_cell_after
        self.log1p = log1p
        return

    def __call__(
        self, 
        sample: dict,
    ) -> dict:
        '''Perform library size normalization in-place
        on a sample dict.

        Parameters
        ----------
        sample : dict
            'input' - torch.FloatTensor, input vector [N, C]
            'output' - torch.LongTensor, target label [N,]

        Returns
        -------
        sample : dict
            'input' - torch.FloatTensor, input vector [N, C]
            'output' - torch.LongTensor, target label [N,]
        '''
        input_ = sample['input']
        size = torch.sum(input_, dim=1).reshape(-1, 1)

        # get proportions of each feature per sample,
        # scale by `counts_per_cell_after`
        prop_input_ = input_ / size
        norm_input_ = prop_input_ * self.counts_per_cell_after
        if self.log1p:
            norm_input_ = torch.log1p(norm_input_)
        sample['input'] = norm_input_
        return sample
    
    
class ExpMinusOne(object):
    
    def __init__(
        self,
    ) -> None:
        '''Perform an exponential minus one transformation
        on an input vector'''
        return
    
    def __call__(
        self,
        sample: dict,
    ) -> dict:
        '''Perform an exponential minus one transformation
        on the sample input.'''
        sample['input'] = torch.expm1(
            sample['input'],
        )
        return sample

    
class MultinomialSample(object):
    '''Sample an mRNA abundance profile from a multinomial
    distribution parameterized by observations.
    '''
    
    def __init__(
        self,
        depth: tuple=(10000, 100000),
        depth_ratio: tuple=None,
    ) -> None:
        '''Sample an mRNA abundance profile from a multinomial 
        distribution parameterized by observations.
        
        Parameters
        ----------
        depth : tuple
            (min, max) depth for multinomial sampling.
        depth_ratio : tuple
            (min, max) ratio of profile depth for multinomial
            sampling. supercedes `depth`.
        
        Returns
        -------
        None.
        '''
        self.depth = depth
        self.depth_ratio = depth_ratio
        
        if self.depth_ratio is not None:
            self.depth = None
        
        return
    
    def __call__(
        self,
        sample: dict,
    ) -> dict:
        '''
        Sample an mRNA profile from a multinomial
        parameterized by observations.
        
        Parameters
        ----------
        sample : dict
            'input' - torch.FloatTensor, input vector [N, C]
            'output' - torch.LongTensor, target label [N,]

        Returns
        -------
        sample : dict
            'input' - torch.FloatTensor, input vector [N, C]
            'output' - torch.LongTensor, target label [N,]
            
        Notes
        -----
        We perform multinomial sampling with a call to `np.random.multinomial`
        for each observation. This may be faster in the future using the native
        `torch.distributions.Multinomial`, but right now the sampling procedure
        is incredibly slow. The implementation below is ~100X slower than our 
        `numpy` calls.
        
            ```
            multi = torch.distributions.Multinomial(
              total_count=d,
              probs=p,
            )
        
            m = multi.sample()
            m = m.float()
            ```
            
        Follow:
        https://github.com/pytorch/pytorch/issues/11931
        '''
        # input is a torch.FloatTensor
        # we assume x is NOT log-transformed
        # cast to float64 to preserve precision of proportions
        x = sample['input'].to(torch.float64)
        size = torch.sum(x, dim=1).detach().cpu().numpy()
        
        # generate a relative abundance profile
        p = (x / torch.sum(x, dim=1).reshape(-1, 1))
        # normalize to ensure roundoff errors don't
        # give us p.sum() > 1
        idx = torch.where(p.sum(1) > 1)
        for i in idx[0]:
            p[i, :] = p[i, :] / np.min([p[i, :].sum(), 1.])
        # sample a sequencing depth
        if self.depth_ratio is None:
            # tile the specified depth for all cells
            depth = np.tile(
                np.array(self.depth).reshape(1, -1),
                (x.size(0), 1)
            ).astype(np.int)
        else:
            # compute a range of depths based on the library size
            # of each observation
            depth = np.concatenate(
                [
                    np.floor(self.depth_ratio[0] * size).reshape(-1, 1),
                    np.ceil(self.depth_ratio[1] * size).reshape(-1, 1),
                ],
                axis = 1,
            ).astype(np.int)
        
        # sample from a multinomial
        # np.random.multinomial is ~100X faster than the native
        # torch.distributions.Multinomial, implemented in Notes
        m = np.zeros(x.size())
        for i in range(x.size(0)):
            
            d = int(
                np.random.choice(
                    np.arange(depth[i, 0], depth[i, 1]),
                    size=1,
                )
            ) 
            
            m[i, :] = np.random.multinomial(
                d,
                pvals=p[i, :].detach().cpu().numpy(),
            )
        m = torch.from_numpy(m).float()
        m = m.to(device=x.device)
        output = {
            'input'  : m,
            'output' : sample['output'],
        }
        return output
    
    
class GeneMasking(object):
    
    def __init__(
        self,
        p_drop: float=0.1,
        p_apply: float=0.5,
        sample_p_drop: bool=False,
    ) -> None:
        '''Mask a subset of genes in the gene expression vector
        with zeros. This may simulate a failed detection event.
        This mask is applied to `p_apply`*100% of input vectors.
        
        Parameters
        ----------
        p_drop : float
            proportion of genes to mask with zeros.
        p_apply : float
            proportion of samples to mask.
        sample_p_drop : bool
            sample the proportion of genes to drop from
            `Unif(0, p_drop)`.
        
        Returns
        -------
        None.
        '''
        self.p_drop = p_drop
        self.p_apply = p_apply
        self.sample_p_drop = sample_p_drop
        return

    def __call__(
        self,
        sample: dict,
    ) -> dict:
        '''Mask a subset of genes.'''
        do_apply = np.random.random()
        if do_apply > self.p_apply:
            # no-op
            return sample
        
        # input is a torch.FloatTensor
        x = sample['input'].clone()
        
        if self.sample_p_drop:
            p_drop = np.random.random() * self.p_drop
        else:
            p_drop = self.p_drop
        
        # mask a proportion `p` of genes with `0`
        # assume x [N, Genes]
        n_genes = x.size(1)
        for i in range(x.size(0)):
            idx = np.random.choice(
                np.arange(n_genes),
                size=int(np.floor(n_genes*p_drop)),
                replace=False,
            ).astype(np.int)
            x[i, idx] = 0
        
        sample['input'] = x
        return sample
    

class InputDropout(object):
    
    def __init__(
        self,
        p_drop: float=0.1,
    ) -> None:
        '''Randomly mask `p_drop` genes.
        
        Parameters
        ----------
        p_drop : float
            proportion of genes to mask.
            
        Returns
        -------
        None
        '''
        self.p_drop = p_drop
        return
    
    def __call__(
        self,
        sample: dict,
    ) -> dict:
        sample['input'] = torch.nn.functional.dropout(
            sample['input'], 
            p=self.p_drop, 
            inplace=False,
        )
        return sample
    
    
class PoissonSample(object):
    '''Sample a gene expression profile based on gene-specific
    Poisson distributions'''
    
    def __init__(
        self,
        depth_range: tuple=(0.5, 2.),
    ) -> None:
        '''Sample a gene expression profile based on gene-specific
        Poisson distributions.
        
        Parameters
        ----------
        depth_range : tuple
            (min_factor, max_factor) for scaling the rate of the Poisson
            that samples are drawn from. Scaling down produces sparser 
            profiles, scaling up produces less sparse profiles.
        
        Returns
        -------
        None.
        
        Notes
        -----
        Treats a raw gene count as an estimate of the rate for a Poisson
        distribution.
        '''
        self.depth_range = depth_range
        return
    
    def __call__(
        self,
        sample: dict,
    ) -> dict:
        # input is a torch.FloatTensor
        # we assume x is NOT log-transformed
        x = sample['input'].to(torch.float64)
        # sample a scale factor for the rate in the specified interval
        # Unif(r1, r2) = Unif(0, 1) * (r1 - r2) + r2
        r = torch.rand(x.size(0)).to(device=x.device)
        r = r * (self.depth_range[0] - self.depth_range[1]) + self.depth_range[1]
        
        P = torch.distributions.Poisson(
            rate=x*r.view(-1, 1),
        )
        x_poisson = P.sample()
        
        assert x.size() == x_poisson.size()
        
        sample['input'] = x_poisson.float()
        return sample


'''Implement MixUp training'''


def mixup(
    a: torch.FloatTensor, 
    b: torch.FloatTensor,
    gamma: torch.FloatTensor,
) -> torch.FloatTensor:
    '''Perform a MixUp operation.
    This is effectively just a weighted average, where
    `gamma = 0.5` yields the mean of `a` and `b`.

    Parameters
    ----------
    a : torch.FloatTensor
        [Batch, C] first sample matrix.
    b : torch.FloatTensor
        [Batch, C] second sample matrix.
    gamma : torch.FloatTensor
        [Batch,] MixUp coefficient.
        
    Returns
    -------
    m : torch.FloatTensor
        [Batch, C] mixed sample matrix.
    '''
    return gamma*a + (1-gamma)*b


class SampleMixUp(object):

    def __init__(
        self, 
        alpha: float=0.2,
        keep_dominant_obs: bool=False,
    ) -> None:
        '''Perform a MixUp operation on a sample batch.

        Parameters
        ----------
        alpha : float
            alpha parameter of the Beta distribution.
        keep_dominant_obs : bool
            use max(gamma, 1-gamma) for each pair of samples
            so the identity of the dominant observation can be
            associated with the mixed sample.

        Returns
        -------
        None.

        References
        ----------
        mixup: Beyond Empirical Risk Minimization
        Hongyi Zhang, Moustapha Cisse, Yann N. Dauphin, David Lopez-Paz
        arXiv:1710.09412

        Notes
        -----
        Zhang et. al. note alpha [0.1, 0.4] improve performance on CIFAR-10,
        while larger values of alpha induce underfitting.
        '''
        self.alpha = alpha
        self.beta = torch.distributions.beta.Beta(
            self.alpha, self.alpha,
        )
        self.keep_dominant_obs = keep_dominant_obs
        return

    def __call__(
        self, 
        sample: dict,
    ) -> dict:
        '''Perform a MixUp operation on the sample.

        Parameters
        ----------
        sample : dict
            'input' - torch.FloatTensor, input vector
            'output' - torch.LongTensor, target label

        Returns
        -------
        sample : dict
            'input' - torch.FloatTensor, input vector
            'output' - torch.LongTensor, target label        
        '''
        if self.alpha == 0.:
            # mixup is deactivated, return the original
            # sample without mixing
            return sample
        
        input_ = sample['input']
        output = sample['output']

        # randomly permute the input and output
        ridx = torch.randperm(input_.size(0))
        r_input_ = input_[ridx]
        r_output = output[ridx]

        # perform the mixup operation between the source
        # data and the rearranged data -- random pairs
        gamma = self.beta.sample((input_.size(0),))
        if self.keep_dominant_obs:
            gamma, _ = torch.max(
                torch.stack([gamma, 1 - gamma,], dim=1,),
                dim=1,
            )
        gamma = gamma.reshape(-1, 1)
        # move gamma weights to the same device as the 
        # inputs
        gamma = gamma.to(device=input_.device)
        
        mix_input_ = mixup(input_, r_input_, gamma=gamma)
        mix_output = mixup(output, r_output, gamma=gamma)

        sample['input'] = mix_input_
        sample['output'] = mix_output
        
        # if there are additional tensors in sample, also mix
        # them up
        other_keys = [
            k for k in sample.keys() if k not in ('input', 'output')
        ]
        for k in other_keys:
            if type(sample[k]) == torch.Tensor:
                sample[k] = mixup(sample[k], sample[k][ridx], gamma=gamma)
        
        # add the randomization index to the sample in case
        # it's useful downstream
        sample['random_idx'] = ridx
        
        return sample


#################################################
# Define augmentation series
#################################################

from torchvision import transforms

def identity(x: Any) -> Any:
    '''Identity function'''
    return x

AUGMENTATION_SCHEMES = {
    'log1p_drop' : transforms.Compose([
        ExpMinusOne(),
        InputDropout(p_drop=0.1,),
        LibrarySizeNormalize(log1p=True),
    ]),
    'log1p_mask' : transforms.Compose([
        ExpMinusOne(),
        GeneMasking(
            p_drop=0.1,
            p_apply=0.5,
        ),
        LibrarySizeNormalize(log1p=True),
    ]),
    'log1p_poisson' : transforms.Compose([
        ExpMinusOne(),
        PoissonSample(),
        LibrarySizeNormalize(log1p=True),
    ]),
    'log1p_poisson_drop' : transforms.Compose([
        ExpMinusOne(),
        PoissonSample(depth_range=(0.1, 2.)),
        InputDropout(p_drop=0.1),
        LibrarySizeNormalize(log1p=True),
    ]),
    'None' : identity,
    'none' : identity,
    None : identity,
}
